@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gsflip.h,v 1.3 2002/04/23 11:58:44 easysw Exp $ */
+/* $Id: gsflip.h,v 1.5 2002/06/16 08:45:42 lpd Exp $ */
 /* Interface to routines for "flipping" image data */
 
 #ifndef gsflip_INCLUDED
@@ -29,8 +29,8 @@
  * bits_per_sample must be 1, 2, 4, 8, or 12.  Returns -1 if num_planes or
  * bits_per_sample is invalid, otherwise 0.
  */
-extern int image_flip_planes(P6(byte * buffer, const byte ** planes,
-				int offset, int nbytes,
-				int num_planes, int bits_per_sample));
+extern int image_flip_planes(byte * buffer, const byte ** planes,
+			     int offset, int nbytes,
+			     int num_planes, int bits_per_sample);
 
 #endif /* gsflip_INCLUDED */
