@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gstype42.c,v 1.4 2002/04/23 11:58:44 easysw Exp $ */
+/* $Id: gstype42.c,v 1.30.2.6 2004/11/15 20:10:27 ray Exp $ */
 /* Type 42 (TrueType) font library routines */
 #include "memory_.h"
 #include "gx.h"
@@ -29,18 +29,27 @@
 #include "gxpath.h"
 #include "gxfont.h"
 #include "gxfont42.h"
+#include "gxttf.h"
+#include "gxttfb.h"
+#include "gxfcache.h"
 #include "gxistate.h"
+#include "stream.h"
 
 /* Structure descriptor */
 public_st_gs_font_type42();
 
 /* Forward references */
-private int append_outline(P4(uint glyph_index, const gs_matrix_fixed * pmat,
-			      gx_path * ppath, gs_font_type42 * pfont));
-private int default_get_outline(P3(gs_font_type42 *pfont, uint glyph_index,
-				   gs_const_string *pgstr));
-private int default_get_metrics(P4(gs_font_type42 *pfont, uint glyph_index,
-				   int wmode, float sbw[4]));
+#if NEW_TT_INTERPRETER
+private int append_outline_fitted(uint glyph_index, const gs_matrix * pmat,
+	       gx_path * ppath, cached_fm_pair * pair, 
+	       const gs_log2_scale_point * pscale, bool design_grid);
+#else
+private int append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
+			   gx_path * ppath, gs_font_type42 * pfont);
+#endif
+private uint default_get_glyph_index(gs_font_type42 *pfont, gs_glyph glyph);
+private int default_get_outline(gs_font_type42 *pfont, uint glyph_index,
+				gs_glyph_data_t *pgd);
 
 /* Set up a pointer to a substring of the font data. */
 /* Free variables: pfont, string_proc. */
@@ -48,6 +57,7 @@ private int default_get_metrics(P4(gs_font_type42 *pfont, uint glyph_index,
   BEGIN\
     code = (*string_proc)(pfont, (ulong)(base), length, &vptr);\
     if ( code < 0 ) return code;\
+    if ( code > 0 ) return_error(gs_error_invalidfont);\
   END
 
 /* Get 2- or 4-byte quantities from a table. */
@@ -59,12 +69,15 @@ private int default_get_metrics(P4(gs_font_type42 *pfont, uint glyph_index,
 
 /* ---------------- Font level ---------------- */
 
-/* Initialize the cached values in a Type 42 font. */
-/* Note that this initializes get_outline and the font procedures as well. */
+/*
+ * Initialize the cached values in a Type 42 font.
+ * Note that this initializes the type42_data procedures other than
+ * string_proc, and the font procedures as well.
+ */
 int
 gs_type42_font_init(gs_font_type42 * pfont)
 {
-    int (*string_proc)(P4(gs_font_type42 *, ulong, uint, const byte **)) =
+    int (*string_proc)(gs_font_type42 *, ulong, uint, const byte **) =
 	pfont->data.string_proc;
     const byte *OffsetTable;
     uint numTables;
@@ -85,13 +98,21 @@ gs_type42_font_init(gs_font_type42 * pfont)
     }
     numTables = U16(OffsetTable + 4);
     ACCESS(12, numTables * 16, TableDirectory);
-    /* Clear optional entries. */
-    memset(pfont->data.metrics, 0, sizeof(pfont->data.metrics));
+    /* Clear all non-client-supplied data. */
+    {
+	void *proc_data = pfont->data.proc_data;
+
+	memset(&pfont->data, 0, sizeof(pfont->data));
+	pfont->data.string_proc = string_proc;
+	pfont->data.proc_data = proc_data;
+    }
     for (i = 0; i < numTables; ++i) {
 	const byte *tab = TableDirectory + i * 16;
 	ulong offset = u32(tab + 8);
 
-	if (!memcmp(tab, "glyf", 4))
+	if (!memcmp(tab, "cmap", 4))
+	    pfont->data.cmap = offset;
+	else if (!memcmp(tab, "glyf", 4))
 	    pfont->data.glyf = offset;
 	else if (!memcmp(tab, "head", 4)) {
 	    const byte *head;
@@ -111,6 +132,11 @@ gs_type42_font_init(gs_font_type42 * pfont)
 	} else if (!memcmp(tab, "loca", 4)) {
 	    pfont->data.loca = offset;
 	    loca_size = u32(tab + 12);
+	} else if (!memcmp(tab, "maxp", 4)) {
+	    const byte *maxp;
+
+	    ACCESS(offset, 30, maxp);
+	    pfont->data.trueNumGlyphs = U16(maxp + 4);
 	} else if (!memcmp(tab, "vhea", 4)) {
 	    const byte *vhea;
 
@@ -120,6 +146,45 @@ gs_type42_font_init(gs_font_type42 * pfont)
 	    pfont->data.metrics[1].offset = offset;
 	    pfont->data.metrics[1].length = (uint)u32(tab + 12);
 	}
+#if 1 /* XXXX */
+	else if (!memcmp(tab, "mort", 4)) {
+	    /* find vertical substitution */
+	    byte *mort;
+
+	    ACCESS(offset, 20, mort);	/* length not accurate */
+	    if ((uint)u32(mort) == 0x00010000) {
+		int nChains, i;
+		ulong Chainp = 8;
+
+		nChains = (int)u32(mort+4);
+		for (i=1; i <= nChains; i++) {
+		  int nSub, j;
+		  ulong Subp;
+
+		  nSub = U16(mort+Chainp+10);
+		  Subp = Chainp + U16(mort+Chainp+8)*12 + 12;
+		  for (j=1; j <= nSub; j++) {
+		     if ((U16(mort+Subp+2)&7) == 4) {
+			if ((U16(mort+Subp+8) == 6) &&
+			    (U16(mort+Subp+10) == 4)) {
+			  byte *ncgsv;
+			  int  cnt;
+			  cnt = U16(mort+Subp+12);
+			  ACCESS(offset+Subp+20, cnt*4, ncgsv);
+			  if (!pfont->data.mort_ncgsv_off) {
+			    pfont->data.mort_ncgsv_off = offset+Subp+20;
+			    pfont->data.mort_ncgsv_len = cnt*4;
+			  } else
+			    dlprintf("multiple ncgsv ignored.\n");
+			}
+		     }
+		     Subp += U16(mort+Subp);
+		  }
+		  Chainp += (int)u32(mort+Chainp+4);
+		}
+	    }
+	}
+#endif
     }
     loca_size >>= pfont->data.indexToLocFormat + 1;
     pfont->data.numGlyphs = (loca_size == 0 ? 0 : loca_size - 1);
@@ -134,15 +199,34 @@ gs_type42_font_init(gs_font_type42 * pfont)
 	pfont->FontBBox.p.x < -0.5 || pfont->FontBBox.p.x > 0.5 ||
 	pfont->FontBBox.p.y < -0.5 || pfont->FontBBox.p.y > 0.5
 	) {
-	float upem = pfont->data.unitsPerEm;
+	float upem = (float)pfont->data.unitsPerEm;
 
 	pfont->FontBBox.p.x = S16(head_box) / upem;
 	pfont->FontBBox.p.y = S16(head_box + 2) / upem;
 	pfont->FontBBox.q.x = S16(head_box + 4) / upem;
 	pfont->FontBBox.q.y = S16(head_box + 6) / upem;
     }
+#if 1 /* XXX */
+    if (! pfont->data.metrics[1].offset) {
+	/* no vhea/vmtx
+	 * shrink/expand the height to 1.0
+ 	 */
+	float v;
+	v = - pfont->FontBBox.p.y + pfont->FontBBox.q.y;
+	    pfont->FontBBox.p.y /= v;
+	    pfont->FontBBox.q.y /= v;
+
+	    pfont->FontBBox.p.x /= v;
+	    pfont->FontBBox.q.x /= v;
+    }
+#endif
+#if NEW_TT_INTERPRETER
+    pfont->data.warning_patented = false;
+    pfont->data.warning_bad_instruction = false;
+#endif
+    pfont->data.get_glyph_index = default_get_glyph_index;
     pfont->data.get_outline = default_get_outline;
-    pfont->data.get_metrics = default_get_metrics;
+    pfont->data.get_metrics = gs_type42_default_get_metrics;
     pfont->procs.glyph_outline = gs_type42_glyph_outline;
     pfont->procs.glyph_info = gs_type42_glyph_info;
     pfont->procs.enumerate_glyph = gs_type42_enumerate_glyph;
@@ -151,16 +235,6 @@ gs_type42_font_init(gs_font_type42 * pfont)
 
 /* ---------------- Glyph level ---------------- */
 
-/* Define the bits in the component glyph flags. */
-#define cg_argsAreWords 1
-#define cg_argsAreXYValues 2
-#define cg_roundXYToGrid 4
-#define cg_haveScale 8
-#define cg_moreComponents 32
-#define cg_haveXYScale 64
-#define cg_have2x2 128
-#define cg_useMyMetrics 512
-
 /*
  * Parse the definition of one component of a composite glyph.  We don't
  * bother to parse the component index, since the caller can do this so
@@ -180,15 +254,15 @@ parse_component(const byte **pdata, uint *pflags, gs_matrix_fixed *psmat,
     flags = U16(gdata);
     gdata += 4;
     mat = *pmat;
-    if (flags & cg_argsAreXYValues) {
+    if (flags & TT_CG_ARGS_ARE_XY_VALUES) {
 	int arg1, arg2;
 	gs_fixed_point pt;
 
-	if (flags & cg_argsAreWords)
+	if (flags & TT_CG_ARGS_ARE_WORDS)
 	    arg1 = S16(gdata), arg2 = S16(gdata + 2), gdata += 4;
 	else
 	    arg1 = S8(gdata), arg2 = S8(gdata + 1), gdata += 2;
-	if (flags & cg_roundXYToGrid) {
+	if (flags & TT_CG_ROUND_XY_TO_GRID) {
 	    /* We should do something here, but we don't. */
 	}
 	gs_point_transform2fixed(pmat, arg1 * factor,
@@ -199,7 +273,7 @@ parse_component(const byte **pdata, uint *pflags, gs_matrix_fixed *psmat,
 	if (pmp)
 	    pmp[0] = pmp[1] = -1;
     } else {
-	if (flags & cg_argsAreWords) {
+	if (flags & TT_CG_ARGS_ARE_WORDS) {
 	    if (pmp)
 		pmp[0] = U16(gdata), pmp[1] = S16(gdata + 2);
 	    gdata += 4;
@@ -210,16 +284,16 @@ parse_component(const byte **pdata, uint *pflags, gs_matrix_fixed *psmat,
 	}
     }
 #define S2_14(p) (S16(p) / 16384.0)
-    if (flags & cg_haveScale) {
+    if (flags & TT_CG_HAVE_SCALE) {
 	scale_mat.xx = scale_mat.yy = S2_14(gdata);
 	scale_mat.xy = scale_mat.yx = 0;
 	gdata += 2;
-    } else if (flags & cg_haveXYScale) {
+    } else if (flags & TT_CG_HAVE_XY_SCALE) {
 	scale_mat.xx = S2_14(gdata);
 	scale_mat.yy = S2_14(gdata + 2);
 	scale_mat.xy = scale_mat.yx = 0;
 	gdata += 4;
-    } else if (flags & cg_have2x2) {
+    } else if (flags & TT_CG_HAVE_2X2) {
 	scale_mat.xx = S2_14(gdata);
 	scale_mat.xy = S2_14(gdata + 2);
 	scale_mat.yx = S2_14(gdata + 4);
@@ -244,16 +318,19 @@ no_scale:
 private int
 total_points(gs_font_type42 *pfont, uint glyph_index)
 {
-    gs_const_string glyph_string;
+    gs_glyph_data_t glyph_data;
     int code;
-    int ocode = pfont->data.get_outline(pfont, glyph_index, &glyph_string);
-    const byte *gdata = glyph_string.data;
+    int ocode;
+    const byte *gdata;
     int total;
 
+    glyph_data.memory = pfont->memory;
+    ocode = pfont->data.get_outline(pfont, glyph_index, &glyph_data);
     if (ocode < 0)
 	return ocode;
-    if (glyph_string.size == 0)
+    if (glyph_data.bits.size == 0)
 	return 0;
+    gdata = glyph_data.bits.data;
     if (S16(gdata) != -1) {
 	/* This is a simple glyph. */
 	int numContours = S16(gdata);
@@ -276,23 +353,33 @@ total_points(gs_font_type42 *pfont, uint glyph_index)
 	    total += code;
 	    parse_component(&gdata, &flags, &mat, NULL, pfont, &mat);
 	}
-	while (flags & cg_moreComponents);
+	while (flags & TT_CG_MORE_COMPONENTS);
     }
-    if (ocode > 0)
-	gs_free_const_string(pfont->memory, gdata, glyph_string.size,
-			     "total_points");
+    gs_glyph_data_free(&glyph_data, "total_points");
     return total;
 }
 
+/*
+ * Define the default implementation for getting the glyph index from a
+ * gs_glyph.  This is trivial for integer ("CID" but representing a GID)
+ * gs_glyph values, and not implemented for name glyphs.
+ */
+private uint
+default_get_glyph_index(gs_font_type42 *pfont, gs_glyph glyph)
+{
+    return (glyph < GS_MIN_CID_GLYPH ? 0 : /* undefined */
+	    glyph - GS_MIN_CID_GLYPH);
+}
+
 /* Define the default implementation for getting the outline data for */
 /* a glyph, using indexToLocFormat and the loca and glyf tables. */
 /* Set pglyph->data = 0 if the glyph is empty. */
 private int
 default_get_outline(gs_font_type42 * pfont, uint glyph_index,
-		    gs_const_string * pglyph)
+		    gs_glyph_data_t *pgd)
 {
-    int (*string_proc) (P4(gs_font_type42 *, ulong, uint, const byte **)) =
-    pfont->data.string_proc;
+    int (*string_proc) (gs_font_type42 *, ulong, uint, const byte **) =
+	pfont->data.string_proc;
     const byte *ploca;
     ulong glyph_start;
     uint glyph_length;
@@ -314,11 +401,89 @@ default_get_outline(gs_font_type42 * pfont, uint glyph_index,
 	ACCESS(pfont->data.loca + glyph_index * 2 + 2, 2, ploca);
 	glyph_length = ((ulong) U16(ploca) << 1) - glyph_start;
     }
-    pglyph->size = glyph_length;
     if (glyph_length == 0)
-	pglyph->data = 0;
-    else
-	ACCESS(pfont->data.glyf + glyph_start, glyph_length, pglyph->data);
+	gs_glyph_data_from_null(pgd);
+    else {
+	const byte *data;
+	byte *buf;
+
+	code = (*string_proc)(pfont, (ulong)(pfont->data.glyf + glyph_start),
+			      glyph_length, &data);
+	if (code < 0) 
+	    return code;
+	if (code == 0)
+	    gs_glyph_data_from_string(pgd, data, glyph_length, NULL);
+	else {
+	    /*
+	     * The glyph is segmented in sfnts. 
+	     * It is not allowed with Type 42 specification.
+	     * Perhaps we can handle it (with a low performance),
+	     * making a contiguous copy.
+	     */
+	    uint left = glyph_length;
+
+	    /* 'code' is the returned length */
+	    buf = (byte *)gs_alloc_string(pgd->memory, glyph_length, "default_get_outline");
+	    if (buf == 0)
+		return_error(gs_error_VMerror);
+	    gs_glyph_data_from_string(pgd, buf, glyph_length, (gs_font *)pfont);
+	    for (;;) {
+		memcpy(buf + glyph_length - left, data, code);
+		if (!(left -= code))
+		    return 0;
+		code = (*string_proc)(pfont, (ulong)(pfont->data.glyf + glyph_start + 
+		              glyph_length - left), left, &data);
+		if (code < 0) 
+		    return code;
+		if (code == 0) 
+		    code = left;
+	    }
+	}
+    }
+    return 0;
+}
+
+/* Take outline data from a True Type font file. */
+int
+gs_type42_get_outline_from_TT_file(gs_font_type42 * pfont, stream *s, uint glyph_index,
+		gs_glyph_data_t *pgd)
+{
+    uchar lbuf[8];
+    ulong glyph_start;
+    uint glyph_length, count;
+
+    /* fixme: Since this function is being called multiple times per glyph,
+     * we should cache glyph data in a buffer associated with the font.
+     */
+    if (pfont->data.indexToLocFormat) {
+	sseek(s, pfont->data.loca + glyph_index * 4);
+        sgets(s, lbuf, 8, &count);
+	if (count < 8)
+	    return_error(gs_error_invalidfont);
+	glyph_start = u32(lbuf);
+	glyph_length = u32(lbuf + 4) - glyph_start;
+    } else {
+	sseek(s, pfont->data.loca + glyph_index * 2);
+        sgets(s, lbuf, 4, &count);
+	if (count < 4)
+	    return_error(gs_error_invalidfont);
+	glyph_start = (ulong) U16(lbuf) << 1;
+	glyph_length = ((ulong) U16(lbuf + 2) << 1) - glyph_start;
+    }
+    if (glyph_length == 0)
+	gs_glyph_data_from_null(pgd);
+    else {
+	byte *buf;
+
+	sseek(s, pfont->data.glyf + glyph_start);
+	buf = (byte *)gs_alloc_string(pgd->memory, glyph_length, "default_get_outline");
+	if (buf == 0)
+	    return_error(gs_error_VMerror);
+	gs_glyph_data_from_string(pgd, buf, glyph_length, (gs_font *)pfont);
+        sgets(s, buf, glyph_length, &count);
+	if (count < glyph_length)
+	    return_error(gs_error_invalidfont);
+    }
     return 0;
 }
 
@@ -327,71 +492,101 @@ private int
 parse_pieces(gs_font_type42 *pfont, gs_glyph glyph, gs_glyph *pieces,
 	     int *pnum_pieces)
 {
-    uint glyph_index = glyph - gs_min_cid_glyph;
-    gs_const_string glyph_string;
-    int code = pfont->data.get_outline(pfont, glyph_index, &glyph_string);
+    uint glyph_index = (glyph >= GS_MIN_GLYPH_INDEX 
+			? glyph - GS_MIN_GLYPH_INDEX 
+			: pfont->data.get_glyph_index(pfont, glyph));
+    gs_glyph_data_t glyph_data;
+    int code;
 
+    glyph_data.memory = pfont->memory;
+    code = pfont->data.get_outline(pfont, glyph_index, &glyph_data);
     if (code < 0)
 	return code;
-    if (glyph_string.size != 0 && S16(glyph_string.data) == -1) {
+    if (glyph_data.bits.size != 0 && S16(glyph_data.bits.data) == -1) {
 	/* This is a composite glyph. */
 	int i = 0;
-	uint flags = cg_moreComponents;
-	const byte *gdata = glyph_string.data + 10;
+	uint flags = TT_CG_MORE_COMPONENTS;
+	const byte *gdata = glyph_data.bits.data + 10;
 	gs_matrix_fixed mat;
 
 	memset(&mat, 0, sizeof(mat)); /* arbitrary */
-	for (i = 0; flags & cg_moreComponents; ++i) {
+	for (i = 0; flags & TT_CG_MORE_COMPONENTS; ++i) {
 	    if (pieces)
-		pieces[i] = U16(gdata + 2) + gs_min_cid_glyph;
+		pieces[i] = U16(gdata + 2) + GS_MIN_GLYPH_INDEX;
 	    parse_component(&gdata, &flags, &mat, NULL, pfont, &mat);
 	}
 	*pnum_pieces = i;
     } else
 	*pnum_pieces = 0;
-    if (code > 0)
-	gs_free_const_string(pfont->memory, glyph_string.data,
-			     glyph_string.size, "parse_pieces");
+    gs_glyph_data_free(&glyph_data, "parse_pieces");
     return 0;
 }
 
 /* Define the font procedures for a Type 42 font. */
 int
-gs_type42_glyph_outline(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
-			gx_path *ppath)
+gs_type42_glyph_outline(gs_font *font, int WMode, gs_glyph glyph, const gs_matrix *pmat,
+			gx_path *ppath, double sbw[4])
 {
     gs_font_type42 *const pfont = (gs_font_type42 *)font;
-    uint glyph_index = glyph - gs_min_cid_glyph;
+    uint glyph_index = (glyph >= GS_MIN_GLYPH_INDEX 
+		? glyph - GS_MIN_GLYPH_INDEX 
+		: pfont->data.get_glyph_index(pfont, glyph));
     gs_fixed_point origin;
     int code;
     gs_glyph_info_t info;
+#if !NEW_TT_INTERPRETER
     gs_matrix_fixed fmat;
+#endif
     static const gs_matrix imat = { identity_matrix_body };
+#if NEW_TT_INTERPRETER
+    bool design_grid = true;
+    const gs_log2_scale_point log2_scale = {0, 0}; 
+    /* fixme : The subpixel numbers doesn't pass through the font_proc_glyph_outline interface.
+       High level devices can't get a proper grid fitting with AlignToPixels = 1.
+       Currently font_proc_glyph_outline is only used by pdfwrite for computing a
+       character bbox, which doesn't need a grid fitting.
+       We apply design grid here.
+     */
+    cached_fm_pair *pair;
+    code = gx_lookup_fm_pair(font, pmat, &log2_scale, design_grid, &pair);
+
+    if (code < 0)
+	return code;
+#endif
 
     if (pmat == 0)
 	pmat = &imat;
-    if ((code = gs_matrix_fixed_from_matrix(&fmat, pmat)) < 0 ||
-	(code = gx_path_current_point(ppath, &origin)) < 0 ||
+    if ((code = gx_path_current_point(ppath, &origin)) < 0 ||
+#if NEW_TT_INTERPRETER
+	(code = append_outline_fitted(glyph_index, pmat, ppath, pair, 
+					&log2_scale, design_grid)) < 0 ||
+#else
+	(code = gs_matrix_fixed_from_matrix(&fmat, pmat)) < 0 ||
 	(code = append_outline(glyph_index, &fmat, ppath, pfont)) < 0 ||
+	/* fixme : don't call glyph_info. */
+#endif
 	(code = font->procs.glyph_info(font, glyph, pmat,
-				       GLYPH_INFO_WIDTH, &info)) < 0
+				       GLYPH_INFO_WIDTH0 << WMode, &info)) < 0
 	)
 	return code;
-    return gx_path_add_point(ppath, origin.x + float2fixed(info.width[0].x),
-			     origin.y + float2fixed(info.width[0].y));
+    return gx_path_add_point(ppath, origin.x + float2fixed(info.width[WMode].x),
+			     origin.y + float2fixed(info.width[WMode].y));
 }
+
+/* Get glyph info by glyph index. */
 int
-gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
-		     int members, gs_glyph_info_t *info)
+gs_type42_glyph_info_by_gid(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
+		     int members, gs_glyph_info_t *info, uint glyph_index)
 {
     gs_font_type42 *const pfont = (gs_font_type42 *)font;
-    uint glyph_index = glyph - gs_min_cid_glyph;
     int default_members =
 	members & ~(GLYPH_INFO_WIDTHS | GLYPH_INFO_NUM_PIECES |
-		    GLYPH_INFO_PIECES);
-    gs_const_string outline;
+		    GLYPH_INFO_PIECES | GLYPH_INFO_OUTLINE_WIDTHS |
+		    GLYPH_INFO_VVECTOR0 | GLYPH_INFO_VVECTOR1);
+    gs_glyph_data_t outline;
     int code = 0;
 
+    outline.memory = pfont->memory;
     if (default_members) {
 	code = gs_default_glyph_info(font, glyph, pmat, default_members, info);
 
@@ -400,12 +595,10 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
     } else if ((code = pfont->data.get_outline(pfont, glyph_index, &outline)) < 0)
 	return code;		/* non-existent glyph */
     else {
-	if (code > 0)
-	    gs_free_const_string(pfont->memory, outline.data, outline.size,
-				 "gs_type42_glyph_info");
+	gs_glyph_data_free(&outline, "gs_type42_glyph_info");
 	info->members = 0;
     }
-    if (members & GLYPH_INFO_WIDTH) {
+    if (members & GLYPH_INFO_WIDTHS) {
 	int i;
 
 	for (i = 0; i < 2; ++i)
@@ -413,15 +606,25 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
 		float sbw[4];
 
 		code = gs_type42_wmode_metrics(pfont, glyph_index, i, sbw);
-		if (code < 0)
-		    return code;
-		if (pmat)
+		if (code < 0) {
+		    code = 0;
+		    continue;
+		}
+		if (pmat) {
 		    code = gs_point_transform(sbw[2], sbw[3], pmat,
 					      &info->width[i]);
-		else
+		    if (code < 0)
+			return code;
+		    code = gs_point_transform(sbw[0], sbw[1], pmat,
+					      &info->v);
+		} else {
 		    info->width[i].x = sbw[2], info->width[i].y = sbw[3];
+		    info->v.x = sbw[0], info->v.y = sbw[1];
+		}
+		info->members |= (GLYPH_INFO_VVECTOR0 << i);
+		info->members |= (GLYPH_INFO_WIDTH << i);
 	    }
-	info->members |= members & GLYPH_INFO_WIDTH;
+	
     }
     if (members & (GLYPH_INFO_NUM_PIECES | GLYPH_INFO_PIECES)) {
 	gs_glyph *pieces =
@@ -435,24 +638,41 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
     return code;
 }
 int
+gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
+		     int members, gs_glyph_info_t *info)
+{
+    gs_font_type42 *const pfont = (gs_font_type42 *)font;
+    uint glyph_index;
+    
+    if (glyph >= GS_MIN_GLYPH_INDEX)
+	glyph_index = glyph - GS_MIN_GLYPH_INDEX;
+    else {
+	glyph_index = pfont->data.get_glyph_index(pfont, glyph);
+	if (glyph_index == GS_NO_GLYPH)
+	    return_error(gs_error_undefined);
+    }
+    return gs_type42_glyph_info_by_gid(font, glyph, pmat, members, info, glyph_index);
+
+}
+int
 gs_type42_enumerate_glyph(gs_font *font, int *pindex,
 			  gs_glyph_space_t glyph_space, gs_glyph *pglyph)
 {
     gs_font_type42 *const pfont = (gs_font_type42 *)font;
 
     while (++*pindex <= pfont->data.numGlyphs) {
-	gs_const_string outline;
+	gs_glyph_data_t outline;
 	uint glyph_index = *pindex - 1;
-	int code = pfont->data.get_outline(pfont, glyph_index, &outline);
+	int code;
 
+	outline.memory = pfont->memory;
+	code = pfont->data.get_outline(pfont, glyph_index, &outline);
 	if (code < 0)
 	    return code;
-	if (outline.data == 0)
+	if (outline.bits.data == 0)
 	    continue;		/* empty (undefined) glyph */
-	*pglyph = glyph_index + gs_min_cid_glyph;
-	if (code > 0)
-	    gs_free_const_string(pfont->memory, outline.data, outline.size,
-				 "gs_type42_enumerate_glyph");
+	*pglyph = glyph_index + GS_MIN_GLYPH_INDEX;
+	gs_glyph_data_free(&outline, "gs_type42_enumerate_glyph");
 	return 0;
     }
     /* We are done. */
@@ -465,7 +685,7 @@ private int
 simple_glyph_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
 		     float sbw[4])
 {
-    int (*string_proc)(P4(gs_font_type42 *, ulong, uint, const byte **)) =
+    int (*string_proc)(gs_font_type42 *, ulong, uint, const byte **) =
 	pfont->data.string_proc;
     double factor = 1.0 / pfont->data.unitsPerEm;
     uint width;
@@ -477,6 +697,8 @@ simple_glyph_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
 	uint num_metrics = pmtx->numMetrics;
 	const byte *pmetrics;
 
+	if (pmtx->length == 0)
+	    return_error(gs_error_rangecheck);
 	if (glyph_index < num_metrics) {
 	    ACCESS(pmtx->offset + glyph_index * 4, 4, pmetrics);
 	    width = U16(pmetrics);
@@ -506,20 +728,22 @@ simple_glyph_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
 }
 
 /* Get the metrics of a glyph. */
-private int
-default_get_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
-		    float sbw[4])
+int
+gs_type42_default_get_metrics(gs_font_type42 * pfont, uint glyph_index,
+			      int wmode, float sbw[4])
 {
-    gs_const_string glyph_string;
-    int code = pfont->data.get_outline(pfont, glyph_index, &glyph_string);
+    gs_glyph_data_t glyph_data;
+    int code;
     int result;
 
+    glyph_data.memory = pfont->memory;
+    code = pfont->data.get_outline(pfont, glyph_index, &glyph_data);
     if (code < 0)
 	return code;
-    if (glyph_string.size != 0 && S16(glyph_string.data) == -1) {
+    if (glyph_data.bits.size != 0 && S16(glyph_data.bits.data) == -1) {
 	/* This is a composite glyph. */
 	uint flags;
-	const byte *gdata = glyph_string.data + 10;
+	const byte *gdata = glyph_data.bits.data + 10;
 	gs_matrix_fixed mat;
 
 	memset(&mat, 0, sizeof(mat)); /* arbitrary */
@@ -527,18 +751,16 @@ default_get_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
 	    uint comp_index = U16(gdata + 2);
 
 	    parse_component(&gdata, &flags, &mat, NULL, pfont, &mat);
-	    if (flags & cg_useMyMetrics) {
+	    if (flags & TT_CG_USE_MY_METRICS) {
 		result = gs_type42_wmode_metrics(pfont, comp_index, wmode, sbw);
 		goto done;
 	    }
 	}
-	while (flags & cg_moreComponents);
+	while (flags & TT_CG_MORE_COMPONENTS);
     }
     result = simple_glyph_metrics(pfont, glyph_index, wmode, sbw);
  done:
-    if (code > 0)
-	gs_free_const_string(pfont->memory, glyph_string.data,
-			     glyph_string.size, "default_get_metrics");
+    gs_glyph_data_free(&glyph_data, "gs_type42_default_get_metrics");
     return result;
 }
 int
@@ -566,6 +788,22 @@ gs_type42_get_metrics(gs_font_type42 * pfont, uint glyph_index,
 
 /* Append a TrueType outline to a path. */
 /* Note that this does not append the final moveto for the width. */
+#if NEW_TT_INTERPRETER
+int
+gs_type42_append(uint glyph_index, gs_imager_state * pis,
+		 gx_path * ppath, const gs_log2_scale_point * pscale,
+		 bool charpath_flag, int paint_type, cached_fm_pair *pair)
+{
+    int code = append_outline_fitted(glyph_index, &ctm_only(pis), ppath, 
+			pair, pscale, charpath_flag);
+
+    if (code < 0)
+	return code;
+    /* Set the flatness for curve rendering. */
+    return gs_imager_setflat(pis, gs_char_flatness(pis, 1.0));
+}
+
+#else
 int
 gs_type42_append(uint glyph_index, gs_imager_state * pis,
 		 gx_path * ppath, const gs_log2_scale_point * pscale,
@@ -578,6 +816,8 @@ gs_type42_append(uint glyph_index, gs_imager_state * pis,
     /* Set the flatness for curve rendering. */
     return gs_imager_setflat(pis, gs_char_flatness(pis, 1.0));
 }
+#endif
+
 
 /* Add 2nd degree Bezier to the path */
 private int
@@ -595,7 +835,8 @@ add_quadratic_curve(gx_path * const ppath, const gs_fixed_point * const a,
  */
 private int
 append_simple(const byte *gdata, float sbw[4], const gs_matrix_fixed *pmat,
-	      gx_path *ppath, gs_fixed_point *ppts, gs_font_type42 * pfont)
+	      gx_path *ppath, gs_fixed_point *ppts, gs_font_type42 * pfont,
+	      bool subglyph)
 {
     int numContours = S16(gdata);
     const byte *pends = gdata + 10;
@@ -637,6 +878,7 @@ append_simple(const byte *gdata, float sbw[4], const gs_matrix_fixed *pmat,
 
     {
 	uint i, np;
+	float offset = 0;
 	gs_fixed_point pt;
 	double factor = 1.0 / pfont->data.unitsPerEm;
 	/*
@@ -646,13 +888,12 @@ append_simple(const byte *gdata, float sbw[4], const gs_matrix_fixed *pmat,
 	byte flags = *pflags++;
 	uint reps = (flags & gf_Repeat ? *pflags++ + 1 : 1);
 
-	/*
-	 * The TrueType documentation gives no clue as to how the lsb
-	 * should affect placement of the outline.  Our best guess is
-	 * that the outline should be translated by lsb - xMin.
-	 */
-	gs_point_transform2fixed(pmat, sbw[0] - S16(gdata + 2) * factor,
-				 0.0, &pt);
+	if (!subglyph) {
+	    int xmin = S16(gdata + 2); /* We like to see it with debugger. */
+
+	    offset = sbw[0] - xmin * factor;
+	}
+	gs_point_transform2fixed(pmat, offset, 0.0, &pt);
 	for (i = 0, np = 0; i < numContours; ++i) {
 	    bool move = true;
 	    bool off_curve = false;
@@ -789,49 +1030,44 @@ append_simple(const byte *gdata, float sbw[4], const gs_matrix_fixed *pmat,
 private int
 check_component(uint glyph_index, const gs_matrix_fixed *pmat,
 		gx_path *ppath, gs_font_type42 *pfont, gs_fixed_point *ppts,
-		gs_const_string *pgstr, bool *pfree_data)
+		gs_glyph_data_t *pgd, bool subglyph)
 {
-    gs_const_string glyph_string;
+    gs_glyph_data_t glyph_data;
     const byte *gdata;
     float sbw[4];
     int numContours;
-    int ocode, code;
+    int code;
 
-    ocode = pfont->data.get_outline(pfont, glyph_index, &glyph_string);
-    if (ocode < 0)
-	return ocode;
-    gdata = glyph_string.data;
-    if (gdata == 0 || glyph_string.size == 0)	/* empty glyph */
+    glyph_data.memory = pfont->memory;
+    code = pfont->data.get_outline(pfont, glyph_index, &glyph_data);
+    if (code < 0)
+	return code;
+    gdata = glyph_data.bits.data;
+    if (gdata == 0 || glyph_data.bits.size == 0)	/* empty glyph */
 	return 0;
     numContours = S16(gdata);
     if (numContours >= 0) {
-        /* Only sbw[0] is used in append_simple. 
-	   In order to set sbw[0], 0 is passed to 
-	   simple_glyph_metrics as wmode. */
-	simple_glyph_metrics(pfont, glyph_index, 0, sbw);
-	code = append_simple(gdata, sbw, pmat, ppath, ppts, pfont);
-	if (ocode > 0)
-	    gs_free_const_string(pfont->memory, gdata, glyph_string.size,
-				 "check_component");
+	gs_type42_get_metrics(pfont, glyph_index, sbw);
+	code = append_simple(gdata, sbw, pmat, ppath, ppts, pfont, subglyph);
+	gs_glyph_data_free(&glyph_data, "check_component");
 	return (code < 0 ? code : 0); /* simple */
     }
     if (numContours != -1)
 	return_error(gs_error_rangecheck);
-    *pgstr = glyph_string;
-    *pfree_data = ocode > 0;
+    *pgd = glyph_data;
     return 1;			/* composite */
 }
 private int
 append_component(uint glyph_index, const gs_matrix_fixed * pmat,
 		 gx_path * ppath, gs_fixed_point *ppts, int point_index,
-		 gs_font_type42 * pfont)
+		 gs_font_type42 * pfont, bool subglyph)
 {
-    gs_const_string gstr;
-    bool free_data;
+    gs_glyph_data_t glyph_data;
     int code;
 
+    glyph_data.memory = pfont->memory;
     code = check_component(glyph_index, pmat, ppath, pfont, ppts + point_index,
-			   &gstr, &free_data);
+			   &glyph_data, subglyph);
     if (code != 1)
 	return code;
     /*
@@ -841,7 +1077,7 @@ append_component(uint glyph_index, const gs_matrix_fixed * pmat,
      */
     {
 	uint flags;
-	const byte *gdata = gstr.data + 10;
+	const byte *gdata = glyph_data.bits.data + 10;
 
 	do {
 	    uint comp_index = U16(gdata + 2);
@@ -861,7 +1097,7 @@ append_component(uint glyph_index, const gs_matrix_fixed * pmat,
 		gs_fixed_point diff;
 
 		code = append_component(comp_index, &mat, NULL, ppts,
-					point_index, pfont);
+					point_index, pfont, true);
 		if (code < 0)
 		    break;
 		diff.x = pfrom->x - pto->x;
@@ -870,28 +1106,27 @@ append_component(uint glyph_index, const gs_matrix_fixed * pmat,
 		mat.ty = fixed2float(mat.ty_fixed += diff.y);
 	    }
 	    code = append_component(comp_index, &mat, ppath, ppts,
-				    point_index, pfont);
+				    point_index, pfont, true);
 	    if (code < 0)
 		break;
 	    point_index += total_points(pfont, comp_index);
 	}
-	while (flags & cg_moreComponents);
+	while (flags & TT_CG_MORE_COMPONENTS);
     }
-    if (free_data)
-	gs_free_const_string(pfont->memory, gstr.data, gstr.size,
-			     "append_component");
+    gs_glyph_data_free(&glyph_data, "append_component");
     return code;
 }
+
+#if !NEW_TT_INTERPRETER
 private int
 append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
 	       gx_path * ppath, gs_font_type42 * pfont)
 {
-    gs_const_string gstr;
-    bool free_data;
-    int code =
-	check_component(glyph_index, pmat, ppath, pfont, NULL, &gstr,
-			&free_data);
+    gs_glyph_data_t glyph_data;
+    glyph_data.memory = pfont->memory;
+    int code;
 
+    code = check_component(glyph_index, pmat, ppath, pfont, NULL, &glyph_data, false);
     if (code != 1)
 	return code;
     {
@@ -907,7 +1142,7 @@ append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
 	if (num_points <= MAX_STACK_PTS) {
 	    gs_fixed_point pts[MAX_STACK_PTS];
 
-	    code = append_component(glyph_index, pmat, ppath, pts, 0, pfont);
+	    code = append_component(glyph_index, pmat, ppath, pts, 0, pfont, false);
 	} else {
 	    gs_memory_t *mem = pfont->memory; /* any memory will do */
 	    gs_fixed_point *ppts = (gs_fixed_point *)
@@ -918,14 +1153,29 @@ append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
 		code = gs_note_error(gs_error_VMerror);
 	    else {
 		code = append_component(glyph_index, pmat, ppath, ppts, 0,
-					pfont);
+					pfont, false);
 		gs_free_object(mem, ppts, "append_outline");
 	    }
 	}
 #undef MAX_STACK_PTS
     }
-    if (free_data)
-	gs_free_const_string(pfont->memory, gstr.data, gstr.size,
-			     "append_outline");
+    gs_glyph_data_free(&glyph_data, "append_outline");
     return code;
 }
+#else
+private int
+append_outline_fitted(uint glyph_index, const gs_matrix * pmat,
+	       gx_path * ppath, cached_fm_pair * pair, 
+	       const gs_log2_scale_point * pscale, bool design_grid)
+{
+    gs_font_type42 *pfont = (gs_font_type42 *)pair->font;
+    int code;
+
+    gx_ttfReader__set_font(pair->ttr, pfont);
+    code = gx_ttf_outline(pair->ttf, pair->ttr, pfont, (uint)glyph_index, 
+	pmat, pscale, ppath, design_grid);
+    gx_ttfReader__set_font(pair->ttr, NULL);
+    return code;
+}
+#endif
+
